==> fling_units-2.0.1.tar.gz: lib/src/core/precision.dart <==
part of fling_units;

/// Represents the amount of precision for a measurement.
///
/// Because measurements are not perfectly precise, we do not want to imply that
/// we know more than what we actually know to be true. For details, see
/// [Significant Figures](https://en.wikipedia.org/wiki/Significant_figures)
///
/// Measurements such as [Distance] will make use of a [Precision] you provide
/// by giving answers within the bounds of the precision you specify when asked
/// for any measurement value.
///
/// In general, measurements will not be considered "equal" if they have
/// different precision, even if their measured values are the same.
///
/// ```dart
/// Distance.meters(3.14159, precision: Precision(3)) ==
///   Distance.meters(3.14159, precision: Precision(5)); // evaluates to false
/// ```
class Precision {
  /// A single digit precision.
  static const Precision single = Precision._single();

  /// Maximum precision.
  static const Precision max = Precision._max();

  /// This precision, in number of digits.
  final int precision;

  /// Constructs a [Precision] of the specified number of digits.
  ///
  /// Due to Dart language limitations on doubles, the maximum precision is 21
  /// digits. Any attempt to use a higher number will fail with an
  /// [ArgumentError]. Precision below 1 digit is meaningless and will also
  /// result in an [ArgumentError].
  Precision(final this.precision) {
    if (precision < 1 || precision > 21) {
      throw ArgumentError(
          'Precision must be a positive integer between 1 and 21');
    }
  }

  /// Combines two [Precision]s per the "multiplication rule".
  ///
  /// Use this when multiplying or dividing two numbers to determine the correct
  /// output precision. See [Wikipedia on Precision Arithmetic](
  /// https://en.wikipedia.org/wiki/Significant_figures#Arithmetic)
  /// for details.
  Precision.combine(final Precision a, final Precision b)
      : this(math.min(a.precision, b.precision));

  /// Combines two [Precision]s per the "addition rule".
  ///
  /// Use this when adding or subtracting two numbers to determine the correct
  /// output precision. See [Wikipedia on Precision Arithmetic](
  /// https://en.wikipedia.org/wiki/Significant_figures#Arithmetic)
  /// for details.
  static Precision addition<T extends Measurement<T>>(
      final Measurement<T> a, final Measurement<T> b) {
    if (a.isInfinite || b.isInfinite) {
      return Precision._max();
    }
    final precisionA = digitsAfterDecimal(a);
    final precisionB = digitsAfterDecimal(b);
    final beforeDecimal = digitsBeforeDecimal(a._preciseSI() + b._preciseSI());
    return Precision(beforeDecimal + math.min(precisionA, precisionB));
  }

  /// Calculates the number of significant digits after the decimal point.
  ///
  /// This only considers digits that fall within the measurement's precision,
  /// e.g.:
  ///
  /// ```dart
  /// digitsAfterDecimal(meters(1.2345));                           // 4
  /// digitsAfterDecimal(meters(1.2345, precision: Precision(3)));  // 2
  /// ```
  static int digitsAfterDecimal(final Measurement measurement) {
    if (measurement.isInfinite || measurement.isNaN) {
      return 0;
    }
    final string = measurement._preciseSI().toStringAsExponential();
    final locationOfE = string.indexOf('e');
    return math.max(
        measurement.precision -
            int.parse(string.substring(locationOfE + 1)) -
            1,
        0);
  }

  /// Calculates the number of digits (significant or not) before the decimal.
  static int digitsBeforeDecimal(final double number) {
    if (number.isInfinite || number.isNaN) {
      return Precision.max.precision;
    }
    final string = number.toStringAsExponential();
    return math.max(
        int.parse(string.substring(string.indexOf('e') + 1)) + 1, 0);
  }

  /// Interprets the specified number according to this Precision.
  double withPrecision(final double value) =>
      double.parse(value.toStringAsPrecision(precision));

  @override
  bool operator ==(final dynamic other) =>
      other is Precision && other.precision == precision;

  @override
  int get hashCode => precision.hashCode;

  /// Single digit precision.
  const Precision._single() : precision = 1;

  /// Maximum precision.
  const Precision._max() : precision = 21;
}

==> fling_units-2.1.0.tar.gz: lib/src/core/precision.dart <==
part of fling_units;

/// Represents the amount of precision for a measurement.
///
/// Because measurements are not perfectly precise, we do not want to imply that
/// we know more than what we actually know to be true. For details, see
/// [Significant Figures](https://en.wikipedia.org/wiki/Significant_figures)
///
/// Measurements such as [Distance] will make use of a [Precision] you provide
/// by giving answers within the bounds of the precision you specify when asked
/// for any measurement value.
///
/// In general, measurements will not be considered "equal" if they have
/// different precision, even if their measured values are the same.
///
/// ```dart
/// Distance.meters(3.14159, precision: Precision(3)) ==
///   Distance.meters(3.14159, precision: Precision(5)); // evaluates to false
/// ```
class Precision {
  /// A single digit precision.
  static const Precision single = Precision._single();

  /// Maximum precision.
  static const Precision max = Precision._max();

  /// This precision, in number of digits.
  final int precision;

  /// Constructs a [Precision] of the specified number of digits.
  ///
  /// Due to Dart language limitations on doubles, the maximum precision is 21
  /// digits. Any attempt to use a higher number will fail with an
  /// [ArgumentError]. Precision below 1 digit is meaningless and will also
  /// result in an [ArgumentError].
  Precision(this.precision) {
    if (precision < 1 || precision > 21) {
      throw ArgumentError(
          'Precision must be a positive integer between 1 and 21');
    }
  }

  /// Combines two [Precision]s per the "multiplication rule".
  ///
  /// Use this when multiplying or dividing two numbers to determine the correct
  /// output precision. See [Wikipedia on Precision Arithmetic](
  /// https://en.wikipedia.org/wiki/Significant_figures#Arithmetic)
  /// for details.
  Precision.combine(final Precision a, final Precision b)
      : this(math.min(a.precision, b.precision));

  /// Combines two [Precision]s per the "addition rule".
  ///
  /// Use this when adding or subtracting two numbers to determine the correct
  /// output precision. See [Wikipedia on Precision Arithmetic](
  /// https://en.wikipedia.org/wiki/Significant_figures#Arithmetic)
  /// for details.
  static Precision addition<T extends Measurement<T>>(
      final Measurement<T> a, final Measurement<T> b) {
    if (a.isInfinite || b.isInfinite) {
      return Precision._max();
    }
    final precisionA = digitsAfterDecimal(a);
    final precisionB = digitsAfterDecimal(b);
    final beforeDecimal = digitsBeforeDecimal(a._preciseSI() + b._preciseSI());
    return Precision(beforeDecimal + math.min(precisionA, precisionB));
  }

  /// Calculates the number of significant digits after the decimal point.
  ///
  /// This only considers digits that fall within the measurement's precision,
  /// e.g.:
  ///
  /// ```dart
  /// digitsAfterDecimal(meters(1.2345));                           // 4
  /// digitsAfterDecimal(meters(1.2345, precision: Precision(3)));  // 2
  /// ```
  static int digitsAfterDecimal(final Measurement measurement) {
    if (measurement.isInfinite || measurement.isNaN) {
      return 0;
    }
    final string = measurement._preciseSI().toStringAsExponential();
    final locationOfE = string.indexOf('e');
    return math.max(
        measurement.precision -
            int.parse(string.substring(locationOfE + 1)) -
            1,
        0);
  }

  /// Calculates the number of digits (significant or not) before the decimal.
  static int digitsBeforeDecimal(final double number) {
    if (number.isInfinite || number.isNaN) {
      return Precision.max.precision;
    }
    final string = number.toStringAsExponential();
    return math.max(
        int.parse(string.substring(string.indexOf('e') + 1)) + 1, 0);
  }

  /// Interprets the specified number according to this Precision.
  double withPrecision(final double value) =>
      double.parse(value.toStringAsPrecision(precision));

  @override
  bool operator ==(final dynamic other) =>
      other is Precision && other.precision == precision;

  @override
  int get hashCode => precision.hashCode;

  /// Single digit precision.
  const Precision._single() : precision = 1;

  /// Maximum precision.
  const Precision._max() : precision = 21;
}
